"""
Download, transform and simulate various datasets.

These classes were extracted from the `utils.py` script from AlgoWit's
publications repo, to which I have also contributed.

Link to related repo: https://github.com/AlgoWit/publications
"""

from ._base import (
    Datasets,
    ImbalancedBinaryDatasets,
    BinaryDatasets,
    ContinuousCategoricalDatasets,
    MulticlassDatasets,
    RemoteSensingDatasets
)

__all__ = [
    'Datasets',
    'ImbalancedBinaryDatasets',
    'BinaryDatasets',
    'ContinuousCategoricalDatasets',
    'MulticlassDatasets',
    'RemoteSensingDatasets'
]



import numpy as np
from sklearn.metrics import SCORERS, make_scorer
from sklearn.metrics._scorer import _PredictScorer
from imblearn.metrics import geometric_mean_score
from rlearn.model_selection.search import MultiClassifier


class ALScorer(_PredictScorer):
    def __init__(self, score_func, sign=1):
        super().__init__(score_func, sign, {})

    def _score(self, method_caller, estimator, X, y_true, sample_weight=None):
        """Evaluate predicted target values for X relative to y_true.
        Parameters
        ----------
        method_caller : callable
            Returns predictions given an estimator, method name, and other
            arguments, potentially caching results.
        estimator : object
            Trained estimator to use for scoring. Must have a predict_proba
            method; the output of that is used to compute the score.
        X : {array-like, sparse matrix}
            Test data that will be fed to estimator.predict.
        y_true : array-like
            Gold standard target values for X.
        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights.
        Returns
        -------
        score : float
            Score function applied to prediction of estimator on X.
        """

        if type(estimator) == MultiClassifier:
            data_utilization, test_scores = estimator\
                .estimator_\
                ._get_performance_scores()
        else:
            data_utilization, test_scores = estimator\
                ._get_performance_scores()

        return self._sign * self._score_func(test_scores, data_utilization)


def geometric_mean_score_macro(y_true, y_pred):
    """Geometric mean score with macro average."""
    return geometric_mean_score(y_true, y_pred, average='macro')


def area_under_learning_curve(test_scores, *args):
    """Area under the learning curve. Used in Active Learning experiments."""
    auc = np.sum(test_scores) / len(test_scores)
    return auc


def data_utilization_rate(test_scores, data_utilization, threshold=.8):
    """Data Utilization Rate. Used in Active Learning Experiments."""
    indices = np.where(np.array(test_scores) >= threshold)[0]
    arg = (
        indices[0]
        if len(indices) != 0
        else -1
    )
    dur = (
        data_utilization[arg]
        if arg != -1
        else 1
    )
    return dur


SCORERS['geometric_mean_score_macro'] = make_scorer(geometric_mean_score_macro)
SCORERS['area_under_learning_curve'] = ALScorer(area_under_learning_curve)
SCORERS['data_utilization_rate'] = ALScorer(data_utilization_rate)



"""
A wrapper to allow an automated Active Learning procedure for an
experimental environment.
"""

# Author: Joao Fonseca <jpmrfonseca@gmail.com>
# License: MIT

import numpy as np
from sklearn.base import clone
from sklearn.base import ClassifierMixin, BaseEstimator
from sklearn.utils import check_X_y
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier
from imblearn.pipeline import Pipeline
from ..metrics import SCORERS
from ._selection_methods import UNCERTAINTY_FUNCTIONS
from ._init_methods import init_strategy


class ALSimulation(ClassifierMixin, BaseEstimator):
    """
    Class to simulate Active Learning experiments.

    This algorithm is an implementation of an Active Learning framework as
    presented in [1]_. The initialization strategy is WIP.

    Parameters
    ----------
    classifier : classifier object, default=None
        Classifier to be used as Chooser and Predictor, or a pipeline
        containing both the generator and the classifier.

    generator : generator estimator, default=None
        Generator to be used for artificial data generation within Active
        Learning iterations.

    use_sample_weight : bool, default=False
        Pass ``sample_weights`` as a fit parameter to the generator object. Used to
        generate artificial data around samples with high uncertainty. ``sample_weights``
        is an array-like of shape (n_samples,) containing the probabilities (based on
        uncertainty) for selecting a sample as a center point.

    init_clusterer : clusterer estimator, default=None
        WIP

    init_strategy : WIP, default='random'
        WIP

    selection_strategy : function or {'entropy', 'breaking_ties',\
        'random'}, default='entropy'
        Method used to quantify the chooser's uncertainty level. All predefined functions
        are set up so that a higher value means higher uncertainty (higher likelihood of
        selection) and vice-versa. The uncertainty estimate is used to select the
        instances to be added to the labeled/training dataset. Selection strategies may
        be added or changed in the ``UNCERTAINTY_FUNCTIONS`` dictionary.

    max_iter : int, default=None
        Maximum number of iterations allowed. If None, the experiment will run until 100%
        of the dataset is added to the training set.

    n_initial : int, default=.02
        Number of observations to include in the initial training dataset. If
        ``n_initial`` < 1, then the corresponding percentage of the original dataset
        will be used as the initial training set.

    increment : int, default=.02
        Number of observations to be added to the training dataset at each
        iteration. If ``n_initial`` < 1, then the corresponding percentage of the
        original dataset will be used as the initial training set.

    save_classifiers : bool, default=False
        Save classifiers fit at each iteration. These classifiers are stored
        in a list ``self.classifiers_``.

    save_test_scores : bool, default=True
        If ``True``, test scores are saved in the list ``self.test_scores_``.
        Size of the test set is defined with the ``test_size`` parameter.

    auto_load : bool, default=True
        If `True`, the classifier with the best training score is saved in the
        method ``self.classifier_``. It's the classifier object used in the
        ``predict`` method.

    test_size : float or int, default=None
        If float, should be between 0.0 and 1.0 and represent the proportion of
        the dataset to include in the test split. If int, represents the
        absolute number of test samples. If None, the value is set to 0.25.

    evaluation_metric : string, default='accuracy'
        Metric used to calculate the test scores. See
        ``research.metrics`` for info on available
        performance metrics.

    random_state : int, RandomState instance, default=None
        Control the randomization of the algorithm.

        - If int, ``random_state`` is the seed used by the random number
          generator;
        - If ``RandomState`` instance, random_state is the random number
          generator;
        - If ``None``, the random number generator is the ``RandomState``
          instance used by ``np.random``.

    References
    ----------
    .. [1] Fonseca, J., Douzas, G., Bacao, F. (2021). Increasing the
       Effectiveness of Active Learning: Introducing Artificial Data Generation
       in Active Learning for Land Use/Land Cover Classification. Remote
       Sensing, 13(13), 2619. https://doi.org/10.3390/rs13132619
    """

    def __init__(
        self,
        classifier=None,
        generator=None,
        use_sample_weight=False,
        init_clusterer=None,
        init_strategy='random',
        selection_strategy='entropy',
        max_iter=None,
        n_initial=.02,
        increment=.02,
        save_classifiers=False,
        save_test_scores=True,
        auto_load=True,
        test_size=None,
        evaluation_metric='accuracy',
        random_state=None
    ):
        self.classifier = classifier
        self.generator = generator
        self.use_sample_weight = use_sample_weight
        self.init_clusterer = init_clusterer
        self.init_strategy = init_strategy
        self.selection_strategy = selection_strategy
        self.max_iter = max_iter
        self.n_initial = n_initial
        self.increment = increment

        # Used to find the optimal classifier
        self.auto_load = auto_load
        self.test_size = test_size
        self.save_classifiers = save_classifiers
        self.save_test_scores = save_test_scores
        self.evaluation_metric = evaluation_metric

        self.random_state = random_state

    def _check(self, X, y):
        """Set ups simple initialization parameters to run an AL simulation."""

        X, y = check_X_y(X, y)

        if self.evaluation_metric is None:
            self.evaluation_metric_ = SCORERS['accuracy']
        elif type(self.evaluation_metric) == str:
            self.evaluation_metric_ = SCORERS[self.evaluation_metric]
        else:
            self.evaluation_metric_ = self.evaluation_metric

        if self.classifier is None:
            self._classifier = RandomForestClassifier(
                random_state=self.random_state
            )
        else:
            self._classifier = clone(self.classifier)

        if type(self.selection_strategy) == str:
            self.selection_strategy_ = UNCERTAINTY_FUNCTIONS[
                self.selection_strategy
            ]
        else:
            self.selection_strategy_ = self.selection_strategy

        if type(self.use_sample_weight) != bool:
            raise TypeError("``use_sample_weight`` must be of type ``bool``. Got"
                            f" {self.use_sample_weight} instead.")

        self.max_iter_ = self.max_iter \
            if self.max_iter is not None \
            else np.inf

        if self.save_classifiers or self.save_test_scores:
            self.data_utilization_ = []

        if self.save_classifiers:
            self.classifiers_ = []

        if self.save_test_scores:
            self.test_scores_ = []

        if self.auto_load:
            self.classifier_ = None
            self._top_score = 0

        if self.auto_load or self.save_test_scores:
            X, X_test, y, y_test = train_test_split(
                X, y,
                test_size=self.test_size,
                random_state=self.random_state,
                stratify=y
            )
        else:
            X_test, y_test = (None, None)

        if self.n_initial < 1:
            n_initial = int(np.round(self.n_initial*X.shape[0]))
            self.n_initial_ = n_initial if n_initial >= 2 else 2
        else:
            self.n_initial_ = self.n_initial

        if self.increment < 1:
            self.increment_ = int(np.round(self.increment*X.shape[0]))
        else:
            self.increment_ = self.increment

        return X, X_test, y, y_test

    def _get_performance_scores(self):
        data_utilization = [
            i[1] for i in self.data_utilization_
        ]
        test_scores = self.test_scores_
        return data_utilization, test_scores

    def _save_metadata(self, iter_n, classifier, X_test, y_test, selection):
        """Save metadata from a completed iteration."""

        # Get score for current iteration
        if self.save_test_scores or self.auto_load:
            score = self.evaluation_metric_(
                classifier,
                X_test,
                y_test
            )

        # Save classifier
        if self.save_classifiers:
            self.classifiers_.append(classifier)

        # Save test scores
        if self.save_test_scores:
            self.test_scores_.append(score)

            self.data_utilization_.append(
                (selection.sum(), selection.sum()/selection.shape[0])
            )

        # Replace top classifier
        if self.auto_load:
            if score > self._top_score:
                self._top_score = score
                self.classifier_ = classifier
                self.top_score_iter_ = iter_n

    def fit(self, X, y):
        """
        Run an Active Learning procedure from training set (X, y).

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples.

        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels) as integers or strings.

        Returns
        -------
        self : ALWrapper
            Completed Active Learning procedure
        """

        # Original "unlabeled" dataset
        iter_n = 0
        X, X_test, y, y_test = self._check(X, y)
        selection = np.zeros(shape=(X.shape[0])).astype(bool)
        sample_weight = None

        # Supervisor - Get data according to passed initialization method
        self.init_clusterer_, ids = init_strategy(
            X=X,
            y=y,
            n_initial=self.n_initial_,
            clusterer=self.init_clusterer,
            selection_method=self.init_strategy,
            random_state=self.random_state
        )

        selection[ids] = True

        while iter_n < self.max_iter_:

            # Generator + Chooser (in this case chooser==Predictor)
            if self.generator is not None:
                generator = clone(self.generator)
                chooser = clone(self._classifier)

                classifier = Pipeline([
                    ('generator', generator),
                    ('chooser', chooser)
                ])
            else:
                classifier = clone(self._classifier)

            if isinstance(classifier, Pipeline) and self.use_sample_weight:
                generator = classifier.steps[-2][-1]
                classifier.steps[-2] = ('generator', generator)

            # Generate artificial data and train classifier
            if self.use_sample_weight:
                classifier.fit(X[selection], y[selection],
                               generator__sample_weight=sample_weight)

                # Compute the class probabilities of labeled observations
                labeled_ids = np.argwhere(selection).squeeze()
                probabs_labeled = classifier.predict_proba(X[selection])
                probabs_labeled = np.where(probabs_labeled == 0., 1e-10, probabs_labeled)
            else:
                classifier.fit(X[selection], y[selection])

            # Save metadata from current iteration
            self._save_metadata(
                iter_n, classifier, X_test, y_test, selection
            )

            # Compute the class probabilities of unlabeled observations
            unlabeled_ids = np.argwhere(~selection).squeeze()
            probabs = classifier.predict_proba(X[~selection])
            probabs = np.where(probabs == 0., 1e-10, probabs)

            # Calculate uncertainty
            uncertainty = self.selection_strategy_(probabs)
            if self.use_sample_weight:
                uncertainty_labeled = self.selection_strategy_(probabs_labeled)

            # Get data according to passed selection criterion
            if self.selection_strategy != 'random':
                ids = unlabeled_ids[np.argsort(uncertainty)[::-1][:self.increment_]]
            else:
                rng = np.random.RandomState(self.random_state)
                ids = rng.choice(unlabeled_ids, self.increment_, replace=False)

            selection[ids] = True

            # Update sample weights for the following iteration
            if self.use_sample_weight:
                sample_weight = np.zeros(selection.shape)
                sample_weight[labeled_ids] = uncertainty_labeled
                sample_weight[unlabeled_ids] = uncertainty
                sample_weight = sample_weight[selection]

                # Corner case: when there is no uncertainty
                if np.isnan(sample_weight).all():
                    sample_weight = np.ones(sample_weight.shape)

            # keep track of iter_n
            iter_n += 1

            # stop if all examples have been included
            if selection.all():
                break
            elif selection.sum()+self.increment_ > y.shape[0]:
                self.increment_ = y.shape[0] - selection.sum()

        return self

    def load_best_classifier(self, X, y):
        """
        Loads the best classifier in the ``self.classifiers_`` list.

        The best classifier is used in the predict method according to the
        performance metric passed.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The test input samples.

        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels) as integers or strings.

        Returns
        -------
        self : ALWrapper
            Completed Active Learning procedure
        """
        scores = []
        for classifier in self.classifiers_:
            scores.append(
                self.evaluation_metric_(classifier, X, y)
            )

        self.classifier_ = self.classifiers_[np.argmax(scores)]
        return self

    def predict(self, X):
        """
        Predict class or regression value for X.

        For a classification model, the predicted class for each sample in X is
        returned. For a regression model, the predicted value based on X is
        returned.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The test input samples.

        Returns
        -------
        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            The predicted classes, or the predict values.
        """
        return self.classifier_.predict(X)



"""
This submodule contains the code developed for experiments related to Active
Learning.
"""

from ._active_learning import ALSimulation
from ._selection_methods import (
    UNCERTAINTY_FUNCTIONS
)

__all__ = [
    'ALSimulation',
    'UNCERTAINTY_FUNCTIONS'
]



# Author: Georgios Douzas <gdouzas@icloud.com>
#         Joao Fonseca <jpmrfonseca@gmail.com>
# License: MIT

import numpy as np
import pandas as pd
from os.path import join, dirname, abspath


def generate_mean_std_tbl(mean_vals, std_vals):
    """Generate table that combines mean and sem values."""
    index = mean_vals.iloc[:, :2]
    scores = mean_vals.iloc[:, 2:].applymap('{:,.2f}'.format) \
        + r" $\pm$ " + \
        std_vals.iloc[:, 2:].applymap('{:,.2f}'.format)
    tbl = pd.concat([index, scores], axis=1)
    return tbl


def generate_pvalues_tbl(tbl):
    """Format p-values."""
    for name in tbl.dtypes[tbl.dtypes == float].index:
        tbl[name] = tbl[name].apply(lambda pvalue: '%.1e' % pvalue)
    return tbl


def sort_tbl(
    tbl,
    ds_order=None,
    ovrs_order=None,
    clfs_order=None,
    metrics_order=None
):
    """
    Sort tables rows and columns. Mostly used to format results from
    oversampling experiments.
    """
    cols = tbl.columns
    keys = ['Dataset', 'Oversampler', 'Classifier', 'Metric']
    for key, cat in zip(
        keys, (ds_order, ovrs_order, clfs_order, metrics_order)
    ):
        if key in cols:
            tbl[key] = pd.Categorical(tbl[key], categories=cat)
    key_cols = [col for col in cols if col in keys]
    tbl.sort_values(key_cols, inplace=True)
    if ovrs_order is not None and set(ovrs_order).issubset(cols):
        tbl = tbl[key_cols + list(ovrs_order)]
    return tbl


def generate_paths(filepath):
    """
    Generate data, results and analysis paths.
    """
    prefix_path = join(dirname(abspath(filepath)), '..')
    paths = [
        join(prefix_path, name)
        for name in ('data', 'results', 'analysis')
    ]
    return paths


def make_bold(
    row, maximum=True, num_decimals=2, threshold=None, with_sem=False
):
    """
    Make bold the lowest or highest value(s).
    with_sem simply returns an incomplete textbf latex function.
    """
    row = round(row, num_decimals)
    if threshold is None:
        val = row.max() if maximum else row.min()
        mask = (row == val)
    else:
        mask = (row > threshold) if maximum else (row < threshold)
    formatter = '{0:.%sf}' % num_decimals
    row = row.apply(lambda el: formatter.format(el))
    row[mask] = [
        '\\textbf{%s' % formatter.format(v)
        if with_sem else '\\textbf{%s}' % formatter.format(v)
        for v in row[mask].astype(float)
    ]

    # Return mask only if function is being used to generate
    # a table with sem values
    if with_sem:
        return row, mask
    else:
        return row


def generate_mean_std_tbl_bold(
    mean_vals, std_vals, maximum=True, decimals=2, threshold=None
):
    """
    Generate table that combines mean and sem values.
    """
    mean_bold = mean_vals.apply(
        lambda row: make_bold(
            row, maximum, decimals, threshold, with_sem=True
        )[0],
        axis=1
    )
    mask = mean_vals.apply(
        lambda row: make_bold(
            row, maximum, decimals, threshold, with_sem=True
        )[1],
        axis=1
    ).values
    std_bold = np.round(std_vals, decimals).astype(str)
    std_bold = np.where(mask, std_bold+'}', std_bold)
    scores = mean_bold + r" $\pm$ " + std_bold
    return scores



"""
Data I/O utils. Later on I might add other data handling utilities.
"""
from os import listdir
from os.path import isdir, join
import pandas as pd
from sqlite3 import connect


def load_datasets(data_dir, suffix='', target_exists=True, **read_csv_kwargs):
    """Load datasets from sqlite database and/or csv files."""
    assert isdir(data_dir), '`data_dir` must be a directory.'

    # Filter data by suffix
    dat_names = [
        dat for dat in listdir(data_dir)
        if dat.endswith(suffix)
    ]

    # Read data
    datasets = []
    for dat_name in dat_names:
        data_path = join(data_dir, dat_name)

        # Handle csv data
        if dat_name.endswith('.csv'):
            ds = pd.read_csv(data_path, **read_csv_kwargs)
            name = dat_name.replace('.csv', '').replace('_', ' ').upper()
            if target_exists:
                ds = (ds.iloc[:, :-1], ds.iloc[:, -1])
            datasets.append((name, ds))

        # Handle sqlite database
        elif dat_name.endswith('.db'):
            with connect(data_path) as connection:
                datasets_names = [
                    name[0]
                    for name in connection.execute(
                        "SELECT name FROM sqlite_master WHERE type='table';"
                    )
                ]
                for dataset_name in datasets_names:
                    ds = pd.read_sql(
                        f'select * from "{dataset_name}"', connection
                    )
                    if target_exists:
                        ds = (ds.iloc[:, :-1], ds.iloc[:, -1])
                    datasets.append(
                        (dataset_name.replace('_', ' ').upper(), ds)
                    )
    return datasets



"""
Utility functions related to image processing
"""

import numpy as np
import pandas as pd


def img_array_to_pandas(X, y):
    """
    Converts an image as numpy array (with ground truth) to a pandas dataframe
    """
    shp = X.shape
    columns = [i for i in range(shp[-1])]+['target']
    dat = np.concatenate([
        np.moveaxis(X, -1, 0), np.moveaxis(y, -1, 0)
    ], axis=0).reshape((len(columns), shp[0]*shp[1]))
    return pd.DataFrame(data=dat.T, columns=columns)



"""
Uncertainty estimators to be used along with the ALWrapper object. All functions must be
set up so that a higher value means higher uncertainty (higher likelihood of selection)
and vice-versa.
"""
import numpy as np


def breaking_ties(probabilities):
    """Breaking Ties uncertainty measurement. The output is scaled and reversed."""
    probs_sorted = np.sort(probabilities, axis=1)[:, ::-1]
    # The extra minus is redundant but I kept as a highlight of the change in the
    # original formula.
    bt = -(probs_sorted[:, 0] - probs_sorted[:, 1])
    return (bt - bt.min(0)) / (bt.max(0) - bt.min(0))


UNCERTAINTY_FUNCTIONS = {
    'entropy': (
        lambda probabilities: (-probabilities * np.log2(probabilities)).sum(axis=1)
    ),
    'breaking_ties': breaking_ties,
    'random': (
        lambda probabilities: np.ones(probabilities.shape[0]) / probabilities.shape[0]
    )
}



"""
This submodule contains a variety of general utility functions as well as
tools used to format and prepare tables to incorporate into LaTeX code.

Additionally, an expanded (as compared with scikit-learn's) dictionary of
scorers is also provided.

This code was taken from the `utils.py` script from AlgoWit's
publications repo, to which I have also contributed.

Link to related repo: https://github.com/AlgoWit/publications
"""
from ._utils import (
    generate_mean_std_tbl,
    generate_pvalues_tbl,
    sort_tbl,
    generate_paths,
    make_bold,
    generate_mean_std_tbl_bold
)
from ._image import img_array_to_pandas
from ._data import load_datasets
from ._check_pipelines import (
    check_pipelines,
    check_pipelines_wrapper
)
from ._visualization import (
    load_plt_sns_configs,
    val_to_color
)

__all__ = [
    'generate_mean_std_tbl',
    'generate_pvalues_tbl',
    'sort_tbl',
    'generate_paths',
    'make_bold',
    'generate_mean_std_tbl_bold',
    'img_array_to_pandas',
    'load_datasets',
    'check_pipelines',
    'check_pipelines_wrapper',
    'load_plt_sns_configs',
    'val_to_color'
]



from rlearn.utils import check_random_states
from itertools import product
from imblearn.pipeline import Pipeline
from sklearn.base import clone

def check_pipelines(objects_list, random_state, n_runs):
    """Extract estimators and parameters grids."""

    # Create random states
    random_states = check_random_states(random_state, n_runs)

    pipelines = []
    param_grid = []
    for comb in product(*objects_list):
        name = '|'.join([i[0] for i in comb])

        # name, object, grid
        comb = [(nm, ob, grd) for nm, ob, grd in comb if ob is not None]

        pipelines.append((name, Pipeline([(nm, ob) for nm, ob, _ in comb])))

        grids = {'est_name': [name]}
        for obj_name, obj, sub_grid in comb:
            if 'random_state' in obj.get_params().keys():
                grids[f'{name}__{obj_name}__random_state'] = random_states
            for param, values in sub_grid.items():
                grids[f'{name}__{obj_name}__{param}'] = values
        param_grid.append(grids)

    return pipelines, param_grid


def check_pipelines_wrapper(
    objects_list,
    wrapper,
    random_state,
    n_runs,
    wrapped_only=False
):
    wrapper_label = wrapper[0]
    wrapper_obj = wrapper[1]
    wrapper_grid = wrapper[2]

    estimators, param_grids = check_pipelines(
        objects_list, random_state, n_runs
    )

    wrapped_estimators = [
        (
            f'{wrapper_label}|{name}',
            clone(wrapper_obj).set_params(**{'classifier': pipeline})
        )
        for name, pipeline in estimators
    ]

    wrapped_param_grids = [
        {
            'est_name': [f'{wrapper_label}|{d["est_name"][0]}'],
            **{
                k.replace(
                    d["est_name"][0],
                    f'{wrapper_label}|{d["est_name"][0]}__classifier'
                ):v
                for k, v in d.items() if k != 'est_name'},
            **{f'{wrapper_label}|{d["est_name"][0]}__{k}': v
                for k, v in wrapper_grid.items()}
        } for d in param_grids
    ]
    if wrapped_only:
        return wrapped_estimators, wrapped_param_grids
    else:
        return (
            estimators + wrapped_estimators,
            param_grids + wrapped_param_grids
        )



"""
Wrapper for cluster-based initialization methods and random initialization.
"""
import numpy as np
from scipy.special import softmax
from ._selection_methods import UNCERTAINTY_FUNCTIONS


def init_strategy(
    X,
    y,
    n_initial,
    clusterer=None,
    selection_method=None,
    random_state=None
):
    """
    Defaults to random.

    Selection method is only relevant if a clusterer object is passed.
    Possible selection methods:
    - None (default): random selection
    - centroid: Gets observations close to the centroids of
      the clusters.
    -
    """

    unlabeled_ids = np.indices(X.shape[:1]).squeeze()

    # Random selection
    if clusterer is None or selection_method in ['random', None]:
        rng = np.random.RandomState(random_state)
        ids = rng.choice(unlabeled_ids, n_initial, replace=False)
        # There must be at least 2 different initial classes
        if len(np.unique(y[ids])) == 1:
            ids[-1] = rng.choice(unlabeled_ids[y != y[ids][0]], 1, replace=False)

        return None, ids

    # Cluster-based selection
    clusterer.fit(X)

    if hasattr(clusterer, 'predict_proba'):
        # Use probabilities to compute uncertainty
        probs = clusterer.predict_proba(X)
    else:
        # Use cluster distances to compute probabilities
        dist = clusterer.transform(X)
        # The first one is another possible alternative
        # dist_inv = 1 - (dist / np.expand_dims(dist.max(1), 1))
        dist_inv = (np.expand_dims(dist.max(1), 1) / dist) - 1
        probs = softmax(dist_inv, axis=1)

    uncertainty = UNCERTAINTY_FUNCTIONS[selection_method](probs)
    ids = unlabeled_ids[np.argsort(uncertainty)[::-1][:n_initial]]

    return clusterer, ids



"""
Download, transform and simulate various datasets.
"""

# Author: Georgios Douzas <gdouzas@icloud.com>
#         Joao Fonseca <jpmrfonseca@gmail.com>
# License: MIT

from os.path import join
from re import sub
from collections import Counter
from itertools import product
from urllib.parse import urljoin
from string import ascii_lowercase
from zipfile import ZipFile
from io import BytesIO, StringIO
from sqlite3 import connect
from scipy.io import loadmat
import io

from rich.progress import track
import requests
import numpy as np
import pandas as pd
from sklearn.utils import check_X_y
from imblearn.datasets import make_imbalance

from research.utils import img_array_to_pandas

UCI_URL = 'https://archive.ics.uci.edu/ml/machine-learning-databases/'
KEEL_URL = 'http://sci2s.ugr.es/keel/keel-dataset/datasets/imbalanced/'
GIC_URL = 'http://www.ehu.eus/ccwintco/uploads/'
OPENML_URL = 'https://www.openml.org/data/get_csv/'
FETCH_URLS = {
    'breast_tissue': urljoin(UCI_URL, '00192/BreastTissue.xls'),
    'ecoli': urljoin(UCI_URL, 'ecoli/ecoli.data'),
    'eucalyptus': urljoin(OPENML_URL, '3625/dataset_194_eucalyptus.arff'),
    'glass': urljoin(UCI_URL, 'glass/glass.data'),
    'haberman': urljoin(UCI_URL, 'haberman/haberman.data'),
    'heart': urljoin(UCI_URL, 'statlog/heart/heart.dat'),
    'iris': urljoin(UCI_URL, 'iris/bezdekIris.data'),
    'libras': urljoin(UCI_URL, 'libras/movement_libras.data'),
    'liver': urljoin(UCI_URL, 'liver-disorders/bupa.data'),
    'pima': 'https://gist.githubusercontent.com/ktisha/c21e73a1bd1700294ef790c56c8aec1f/raw/819b69b5736821ccee93d05b51de0510bea00294/pima-indians-diabetes.csv',
    'vehicle': urljoin(UCI_URL, 'statlog/vehicle/'),
    'wine': urljoin(UCI_URL, 'wine/wine.data'),
    'new_thyroid_1': urljoin(urljoin(KEEL_URL, 'imb_IRlowerThan9/'), 'new-thyroid1.zip'),
    'new_thyroid_2': urljoin(urljoin(KEEL_URL, 'imb_IRlowerThan9/'), 'new-thyroid2.zip'),
    'cleveland': urljoin(urljoin(KEEL_URL, 'imb_IRhigherThan9p2/'), 'cleveland-0_vs_4.zip'),
    'led': urljoin(urljoin(KEEL_URL, 'imb_IRhigherThan9p2/'), 'led7digit-0-2-4-5-6-7-8-9_vs_1.zip'),
    'page_blocks_1_3': urljoin(urljoin(KEEL_URL, 'imb_IRhigherThan9p1/'), 'page-blocks-1-3_vs_4.zip'),
    'vowel': urljoin(urljoin(KEEL_URL, 'imb_IRhigherThan9p1/'), 'vowel0.zip'),
    'yeast_1': urljoin(urljoin(KEEL_URL, 'imb_IRlowerThan9/'), 'yeast1.zip'),
    'banknote_authentication': urljoin(UCI_URL, '00267/data_banknote_authentication.txt'),
    'arcene': urljoin(UCI_URL, 'arcene/'),
    'audit': urljoin(UCI_URL, '00475/audit_data.zip'),
    'spambase': urljoin(UCI_URL, 'spambase/spambase.data'),
    'parkinsons': urljoin(UCI_URL, 'parkinsons/parkinsons.data'),
    'ionosphere': urljoin(UCI_URL, 'ionosphere/ionosphere.data'),
    'breast_cancer': urljoin(UCI_URL, 'breast-cancer-wisconsin/wdbc.data'),
    'adult': urljoin(UCI_URL, 'adult/adult.data'),
    'abalone': urljoin(UCI_URL, 'abalone/abalone.data'),
    'acute': urljoin(UCI_URL, 'acute/diagnosis.data'),
    'annealing': urljoin(UCI_URL, 'annealing/anneal.data'),
    'census': urljoin(UCI_URL, 'census-income-mld/census-income.data.gz'),
    'contraceptive': urljoin(UCI_URL, 'cmc/cmc.data'),
    'covertype': urljoin(UCI_URL, 'covtype/covtype.data.gz'),
    'credit_approval': urljoin(UCI_URL, 'credit-screening/crx.data'),
    'dermatology': urljoin(UCI_URL, 'dermatology/dermatology.data'),
    'echocardiogram': urljoin(UCI_URL, 'echocardiogram/echocardiogram.data'),
    'flags': urljoin(UCI_URL, 'flags/flag.data'),
    'heart_disease': [urljoin(UCI_URL, 'heart-disease/processed.cleveland.data'), urljoin(UCI_URL, 'heart-disease/processed.hungarian.data'), urljoin(UCI_URL, 'heart-disease/processed.switzerland.data'), urljoin(UCI_URL, 'heart-disease/processed.va.data')],
    'hepatitis': urljoin(UCI_URL, 'hepatitis/hepatitis.data'),
    'german_credit': urljoin(UCI_URL, 'statlog/german/german.data'),
    'thyroid': urljoin(UCI_URL, 'thyroid-disease/thyroid0387.data'),
    'first_order_theorem': urljoin(OPENML_URL, '1587932/phpPbCMyg'),
    'gas_drift': urljoin(OPENML_URL, '1588715/phpbL6t4U'),
    'autouniv_au7': urljoin(OPENML_URL, '1593748/phpmRPvKy'),
    'autouniv_au4': urljoin(OPENML_URL, '1593744/phpiubDlf'),
    'mice_protein': urljoin(OPENML_URL, '17928620/phpchCuL5'),
    'steel_plates': urljoin(OPENML_URL, '18151921/php5s7Ep8'),
    'cardiotocography': urljoin(OPENML_URL, '1593756/phpW0AXSQ'),
    'waveform': urljoin(OPENML_URL, '60/dataset_60_waveform-5000.arff'),
    'volkert': urljoin(OPENML_URL, '19335689/file1c556e3db171.arff'),
    'asp_potassco': urljoin(OPENML_URL, '21377447/file18547f421393.arff'),
    'wine_quality': urljoin(OPENML_URL, '4965268/wine-quality-red.arff'),
    'mfeat_zernike': urljoin(OPENML_URL, '22/dataset_22_mfeat-zernike.arff'),
    'gesture_segmentation': urljoin(OPENML_URL, '1798765/phpYLeydd'),
    'texture': urljoin(OPENML_URL, '4535764/phpBDgUyY'),
    'usps': urljoin(OPENML_URL, '19329737/usps.arff'),
    'japanese_vowels': urljoin(OPENML_URL, '52415/JapaneseVowels.arff'),
    'pendigits': urljoin(OPENML_URL, '32/dataset_32_pendigits.arff'),
    'image_segmentation': urljoin(OPENML_URL, '18151937/phpyM5ND4'),
    'baseball': urljoin(OPENML_URL, '3622/dataset_189_baseball.arff'),
    'indian_pines': [urljoin(GIC_URL,'2/22/Indian_pines.mat'), urljoin(GIC_URL,'c/c4/Indian_pines_gt.mat')],
    'salinas': [urljoin(GIC_URL,'f/f1/Salinas.mat'), urljoin(GIC_URL,'f/fa/Salinas_gt.mat')],
    'salinas_a': [urljoin(GIC_URL,'d/df/SalinasA.mat'), urljoin(GIC_URL,'a/aa/SalinasA_gt.mat')],
    'pavia_centre': [urljoin(GIC_URL,'e/e3/Pavia.mat'), urljoin(GIC_URL,'5/53/Pavia_gt.mat')],
    'pavia_university': [urljoin(GIC_URL,'e/ee/PaviaU.mat'), urljoin(GIC_URL,'5/50/PaviaU_gt.mat')],
    'kennedy_space_center': [urljoin(GIC_URL,'2/26/KSC.mat'), urljoin(GIC_URL,'a/a6/KSC_gt.mat')],
    'botswana': [urljoin(GIC_URL,'7/72/Botswana.mat'), urljoin(GIC_URL,'5/58/Botswana_gt.mat')]
}
RANDOM_STATE = 0


class Datasets:
    """Base class to download and save datasets."""

    def __init__(self, names='all'):
        self.names = names

    @staticmethod
    def _modify_columns(data):
        """Rename and reorder columns of dataframe."""
        X, y = data.drop(columns='target'), data.target
        X.columns = range(len(X.columns))
        return pd.concat([X, y], axis=1)

    def download(self):
        """Download the datasets."""
        if self.names == 'all':
            func_names = [
                func_name for func_name in dir(self) if 'fetch_' in func_name
            ]
        else:
            func_names = [
                f'fetch_{name}'.lower().replace(' ', '_')
                for name in self.names
            ]
        self.content_ = []
        for func_name in track(func_names, description='Datasets'):
            name = func_name.replace('fetch_', '').upper().replace('_', ' ')
            fetch_data = getattr(self, func_name)
            data = self._modify_columns(fetch_data())
            self.content_.append((name, data))
        return self

    def save(self, path, db_name):
        """Save datasets."""
        with connect(join(path, f'{db_name}.db')) as connection:
            for name, data in self.content_:
                data.to_sql(name, connection, index=False, if_exists='replace')


class ImbalancedBinaryDatasets(Datasets):
    """Class to download, transform and save binary class imbalanced
    datasets."""

    MULTIPLICATION_FACTORS = [2, 3]

    @staticmethod
    def _calculate_ratio(multiplication_factor, y):
        """Calculate ratio based on IRs multiplication factor."""
        ratio = Counter(y).copy()
        ratio[1] = int(ratio[1] / multiplication_factor)
        return ratio

    def _make_imbalance(self, data, multiplication_factor):
        """Undersample the minority class."""
        X_columns = [col for col in data.columns if col != 'target']
        X, y = check_X_y(data.loc[:, X_columns], data.target)
        if multiplication_factor > 1.0:
            sampling_strategy = self._calculate_ratio(multiplication_factor, y)
            X, y = make_imbalance(
                X, y, sampling_strategy=sampling_strategy,
                random_state=RANDOM_STATE
            )
        data = pd.DataFrame(np.column_stack((X, y)))
        data.iloc[:, -1] = data.iloc[:, -1].astype(int)
        return data

    def download(self):
        """Download the datasets and append undersampled versions of them."""
        super(ImbalancedBinaryDatasets, self).download()
        undersampled_datasets = []
        for (name, data), factor in list(
            product(self.content_, self.MULTIPLICATION_FACTORS)
        ):
            ratio = self._calculate_ratio(factor, data.target)
            if ratio[1] >= 15:
                data = self._make_imbalance(data, factor)
                undersampled_datasets.append((f'{name} ({factor})', data))
        self.content_ += undersampled_datasets
        return self

    def fetch_breast_tissue(self):
        """Download and transform the Breast Tissue Data Set.
        The minority class is identified as the `car` and `fad`
        labels and the majority class as the rest of the labels.

        http://archive.ics.uci.edu/ml/datasets/breast+tissue
        """
        data = pd.read_excel(FETCH_URLS['breast_tissue'], sheet_name='Data')
        data = data.drop(columns='Case #').rename(columns={'Class': 'target'})
        data['target'] = data['target'].isin(['car', 'fad']).astype(int)
        return data

    def fetch_ecoli(self):
        """Download and transform the Ecoli Data Set.
        The minority class is identified as the `pp` label
        and the majority class as the rest of the labels.

        https://archive.ics.uci.edu/ml/datasets/ecoli
        """
        data = pd.read_csv(
            FETCH_URLS['ecoli'], header=None, delim_whitespace=True
        )
        data = data.drop(columns=0).rename(columns={8: 'target'})
        data['target'] = data['target'].isin(['pp']).astype(int)
        return data

    def fetch_eucalyptus(self):
        """Download and transform the Eucalyptus Data Set.
        The minority class is identified as the `best` label
        and the majority class as the rest of the labels.

        https://www.openml.org/d/188
        """
        data = pd.read_csv(FETCH_URLS['eucalyptus'])
        data = data.iloc[:, -9:].rename(columns={'Utility': 'target'})
        data = data[data != '?'].dropna()
        data['target'] = data['target'].isin(['best']).astype(int)
        return data

    def fetch_glass(self):
        """Download and transform the Glass Identification Data Set.
        The minority class is identified as the `1` label
        and the majority class as the rest of the labels.

        https://archive.ics.uci.edu/ml/datasets/glass+identification
        """
        data = pd.read_csv(FETCH_URLS['glass'], header=None)
        data = data.drop(columns=0).rename(columns={10: 'target'})
        data['target'] = data['target'].isin([1]).astype(int)
        return data

    def fetch_haberman(self):
        """Download and transform the Haberman's Survival Data Set.
        The minority class is identified as the `1` label
        and the majority class as the `0` label.

        https://archive.ics.uci.edu/ml/datasets/Haberman's+Survival
        """
        data = pd.read_csv(FETCH_URLS['haberman'], header=None)
        data.rename(columns={3: 'target'}, inplace=True)
        data['target'] = data['target'].isin([2]).astype(int)
        return data

    def fetch_heart(self):
        """Download and transform the Heart Data Set.
        The minority class is identified as the `2` label
        and the majority class as the `1` label.

        http://archive.ics.uci.edu/ml/datasets/statlog+(heart)
        """
        data = pd.read_csv(
            FETCH_URLS['heart'], header=None, delim_whitespace=True
        )
        data.rename(columns={13: 'target'}, inplace=True)
        data['target'] = data['target'].isin([2]).astype(int)
        return data

    def fetch_iris(self):
        """Download and transform the Iris Data Set.
        The minority class is identified as the `1` label
        and the majority class as the rest of the labels.

        https://archive.ics.uci.edu/ml/datasets/iris
        """
        data = pd.read_csv(FETCH_URLS['iris'], header=None)
        data.rename(columns={4: 'target'}, inplace=True)
        data['target'] = data['target'].isin(['Iris-setosa']).astype(int)
        return data

    def fetch_libras(self):
        """Download and transform the Libras Movement Data Set.
        The minority class is identified as the `1` label
        and the majority class as the rest of the labels.

        https://archive.ics.uci.edu/ml/datasets/Libras+Movement
        """
        data = pd.read_csv(FETCH_URLS['libras'], header=None)
        data.rename(columns={90: 'target'}, inplace=True)
        data['target'] = data['target'].isin([1]).astype(int)
        return data

    def fetch_liver(self):
        """Download and transform the Liver Disorders Data Set.
        The minority class is identified as the `1` label
        and the majority class as the '2' label.

        https://archive.ics.uci.edu/ml/datasets/liver+disorders
        """
        data = pd.read_csv(FETCH_URLS['liver'], header=None)
        data.rename(columns={6: 'target'}, inplace=True)
        data['target'] = data['target'].isin([1]).astype(int)
        return data

    def fetch_pima(self):
        """Download and transform the Pima Indians Diabetes Data Set.
        The minority class is identified as the `1` label
        and the majority class as the '0' label.

        https://www.kaggle.com/uciml/pima-indians-diabetes-database
        """
        data = pd.read_csv(FETCH_URLS['pima'], header=None, skiprows=9)
        data.rename(columns={8: 'target'}, inplace=True)
        return data

    def fetch_vehicle(self):
        """Download and transform the Vehicle Silhouettes Data Set.
        The minority class is identified as the `1` label
        and the majority class as the rest of the labels.

        https://archive.ics.uci.edu/ml/datasets/Statlog+(Vehicle+Silhouettes)
        """
        data = pd.DataFrame()
        for letter in ascii_lowercase[0:9]:
            partial_data = pd.read_csv(
                urljoin(FETCH_URLS['vehicle'], 'xa%s.dat' % letter),
                header=None, delim_whitespace=True
            )
            partial_data = partial_data.rename(columns={18: 'target'})
            partial_data['target'] = partial_data['target']\
                .isin(['van']).astype(int)
            data = data.append(partial_data)
        return data

    def fetch_wine(self):
        """Download and transform the Wine Data Set.
        The minority class is identified as the `2` label
        and the majority class as the rest of the labels.

        https://archive.ics.uci.edu/ml/datasets/wine
        """
        data = pd.read_csv(FETCH_URLS['wine'], header=None)
        data.rename(columns={0: 'target'}, inplace=True)
        data['target'] = data['target'].isin([2]).astype(int)
        return data

    def fetch_new_thyroid_1(self):
        """Download and transform the Thyroid 1 Disease Data Set.
        The minority class is identified as the `positive`
        label and the majority class as the `negative` label.

        http://sci2s.ugr.es/keel/dataset.php?cod=145
        """
        zipped_data = requests.get(FETCH_URLS['new_thyroid_1']).content
        unzipped_data = ZipFile(BytesIO(zipped_data))\
            .read('new-thyroid1.dat')\
            .decode('utf-8')
        data = pd.read_csv(
            StringIO(sub(r'@.+\n+', '', unzipped_data)),
            header=None, sep=', ', engine='python'
        )
        data.rename(columns={5: 'target'}, inplace=True)
        data['target'] = data['target'].isin(['positive']).astype(int)
        return data

    def fetch_new_thyroid_2(self):
        """Download and transform the Thyroid 2 Disease Data Set.
        The minority class is identified as the `positive`
        label and the majority class as the `negative` label.

        http://sci2s.ugr.es/keel/dataset.php?cod=146
        """
        zipped_data = requests.get(FETCH_URLS['new_thyroid_2']).content
        unzipped_data = ZipFile(BytesIO(zipped_data))\
            .read('newthyroid2.dat')\
            .decode('utf-8')
        data = pd.read_csv(
            StringIO(sub(r'@.+\n+', '', unzipped_data)),
            header=None, sep=', ', engine='python'
        )
        data.rename(columns={5: 'target'}, inplace=True)
        data['target'] = data['target'].isin(['positive']).astype(int)
        return data

    def fetch_cleveland(self):
        """Download and transform the Heart Disease Cleveland Data Set.
        The minority class is identified as the `positive` label and
        the majority class as the `negative` label.

        http://sci2s.ugr.es/keel/dataset.php?cod=980
        """
        zipped_data = requests.get(FETCH_URLS['cleveland']).content
        unzipped_data = ZipFile(BytesIO(zipped_data))\
            .read('cleveland-0_vs_4.dat')\
            .decode('utf-8')
        data = pd.read_csv(
            StringIO(sub(r'@.+\n+', '', unzipped_data)),
            header=None
        )
        data.rename(columns={13: 'target'}, inplace=True)
        data['target'] = data['target'].isin(['positive']).astype(int)
        return data

    def fetch_dermatology(self):
        """Download and transform the Dermatology Data Set.
        The minority class is identified as the `positive` label and
        the majority class as the `negative` label.

        http://sci2s.ugr.es/keel/dataset.php?cod=1330
        """
        data = pd.read_csv(
            FETCH_URLS['dermatology'],
            header=None
        )
        data.rename(columns={34: 'target'}, inplace=True)
        data.drop(columns=33, inplace=True)
        data['target'] = data['target'].isin(['positive']).astype(int)
        return data

    def fetch_led(self):
        """Download and transform the LED Display Domain Data Set.
        The minority class is identified as the `positive` label and
        the majority class as the `negative` label.

        http://sci2s.ugr.es/keel/dataset.php?cod=998
        """
        zipped_data = requests.get(FETCH_URLS['led']).content
        unzipped_data = ZipFile(BytesIO(zipped_data))\
            .read('led7digit-0-2-4-5-6-7-8-9_vs_1.dat')\
            .decode('utf-8')
        data = pd.read_csv(
            StringIO(sub(r'@.+\n+', '', unzipped_data)),
            header=None
        )
        data.rename(columns={7: 'target'}, inplace=True)
        data['target'] = data['target'].isin(['positive']).astype(int)
        return data

    def fetch_page_blocks_1_3(self):
        """Download and transform the Page Blocks 1-3 Data Set.
        The minority class is identified as the `positive` label and
        the majority class as the `negative` label.

        http://sci2s.ugr.es/keel/dataset.php?cod=124
        """
        zipped_data = requests.get(FETCH_URLS['page_blocks_1_3']).content
        unzipped_data = ZipFile(BytesIO(zipped_data))\
            .read('page-blocks-1-3_vs_4.dat')\
            .decode('utf-8')
        data = pd.read_csv(
            StringIO(sub(r'@.+\n+', '', unzipped_data)),
            header=None
        )
        data.rename(columns={10: 'target'}, inplace=True)
        data['target'] = data['target'].isin(['positive']).astype(int)
        return data

    def fetch_vowel(self):
        """Download and transform the Vowel Recognition Data Set.
        The minority class is identified as the `positive` label and
        the majority class as the `negative` label.

        http://sci2s.ugr.es/keel/dataset.php?cod=127
        """
        zipped_data = requests.get(FETCH_URLS['vowel']).content
        unzipped_data = ZipFile(BytesIO(zipped_data))\
            .read('vowel0.dat')\
            .decode('utf-8')
        data = pd.read_csv(
            StringIO(sub(r'@.+\n+', '', unzipped_data)),
            header=None
        )
        data.rename(columns={13: 'target'}, inplace=True)
        data['target'] = data['target'].isin([' positive']).astype(int)
        return data

    def fetch_yeast_1(self):
        """Download and transform the Yeast 1 Data Set.
        The minority class is identified as the `positive` label and
        the majority class as the `negative` label.

        http://sci2s.ugr.es/keel/dataset.php?cod=153
        """
        zipped_data = requests.get(FETCH_URLS['yeast_1']).content
        unzipped_data = ZipFile(BytesIO(zipped_data))\
            .read('yeast1.dat')\
            .decode('utf-8')
        data = pd.read_csv(
            StringIO(sub(r'@.+\n+', '', unzipped_data)),
            header=None
        )
        data.rename(columns={8: 'target'}, inplace=True)
        data['target'] = data['target'].isin([' positive']).astype(int)
        return data


class BinaryDatasets(Datasets):
    """Class to download, transform and save binary class datasets."""

    def fetch_banknote_authentication(self):
        """Download and transform the Banknote Authentication Data Set.

        https://archive.ics.uci.edu/ml/datasets/banknote+authentication
        """
        data = pd.read_csv(FETCH_URLS['banknote_authentication'], header=None)
        data.rename(columns={4: 'target'}, inplace=True)
        return data

    def fetch_arcene(self):
        """Download and transform the Arcene Data Set.

        https://archive.ics.uci.edu/ml/datasets/Arcene
        """
        url = FETCH_URLS['arcene']
        data, labels = [], []
        for data_type in ('train', 'valid'):
            data.append(pd.read_csv(
                urljoin(url, f'ARCENE/arcene_{data_type}.data'),
                header=None, sep=' '
            ).drop(columns=list(range(1998, 10001))))
            labels.append(pd.read_csv(
                urljoin(url, (
                        'ARCENE/' if data_type == 'train' else ''
                        ) + f'arcene_{data_type}.labels'),
                header=None
            ).rename(columns={0: 'target'}))
        data = pd.concat(data, ignore_index=True)
        labels = pd.concat(labels, ignore_index=True)
        data = pd.concat([data, labels], axis=1)
        data['target'] = data['target'].isin([1]).astype(int)
        return data

    def fetch_audit(self):
        """Download and transform the Audit Data Set.

        https://archive.ics.uci.edu/ml/datasets/Audit+Data
        """
        zipped_data = requests.get(FETCH_URLS['audit']).content
        unzipped_data = ZipFile(BytesIO(zipped_data))\
            .read('audit_data/audit_risk.csv')\
            .decode('utf-8')
        data = pd.read_csv(
            StringIO(sub(r'@.+\n+', '', unzipped_data)),
            engine='python'
        )
        data = data\
            .drop(columns=['LOCATION_ID'])\
            .rename(columns={'Risk': 'target'})\
            .dropna()
        return data

    def fetch_spambase(self):
        """Download and transform the Spambase Data Set.

        https://archive.ics.uci.edu/ml/datasets/Spambase
        """
        data = pd.read_csv(FETCH_URLS['spambase'], header=None)
        data.rename(columns={57: 'target'}, inplace=True)
        return data

    def fetch_parkinsons(self):
        """Download and transform the Parkinsons Data Set.

        https://archive.ics.uci.edu/ml/datasets/parkinsons
        """
        data = pd.read_csv(FETCH_URLS['parkinsons'])
        data = pd.concat([
            data.drop(columns=['name', 'status']),
            data[['status']].rename(columns={'status': 'target'})],
            axis=1
        )
        data['target'] = data['target'].isin([0]).astype(int)
        return data

    def fetch_ionosphere(self):
        """Download and transform the Ionosphere Data Set.

        https://archive.ics.uci.edu/ml/datasets/ionosphere
        """
        data = pd.read_csv(FETCH_URLS['ionosphere'], header=None)
        data = data.drop(columns=[0, 1]).rename(columns={34: 'target'})
        data['target'] = data['target'].isin(['b']).astype(int)
        return data

    def fetch_breast_cancer(self):
        """Download and transform the Breast Cancer Wisconsin Data Set.

        https://archive.ics.uci.edu/ml/datasets/Breast+Cancer+Wisconsin+(Diagnostic)
        """
        data = pd.read_csv(FETCH_URLS['breast_cancer'], header=None)
        data = pd.concat([
            data.drop(columns=[0, 1]),
            data[[1]].rename(columns={1: 'target'})],
            axis=1
        )
        data['target'] = data['target'].isin(['M']).astype(int)
        return data


class ContinuousCategoricalDatasets(Datasets):
    """Class to download, transform and save datasets with both continuous
    and categorical features."""

    @staticmethod
    def _modify_columns(data, categorical_features):
        """Rename and reorder columns of dataframe."""
        X, y = data.drop(columns='target'), data.target
        X.columns = range(len(X.columns))
        return pd.concat([X, y], axis=1), categorical_features

    def download(self):
        """Download the datasets."""
        if self.names == 'all':
            func_names = [
                func_name
                for func_name in dir(self)
                if 'fetch_' in func_name
            ]
        else:
            func_names = [
                f'fetch_{name}'.lower().replace(' ', '_')
                for name in self.names
            ]
        self.content_ = []
        for func_name in track(func_names, description='Datasets'):
            name = func_name.replace('fetch_', '').upper().replace('_', ' ')
            fetch_data = getattr(self, func_name)
            data, categorical_features = self._modify_columns(*fetch_data())
            self.content_.append((name, data, categorical_features))
        return self

    def save(self, path, db_name):
        """Save datasets."""
        with connect(join(path, f'{db_name}.db')) as connection:
            for name, data in self.content_:
                data.to_sql(name, connection, index=False, if_exists='replace')

    def fetch_adult(self):
        """Download and transform the Adult Data Set.

        https://archive.ics.uci.edu/ml/datasets/Adult
        """
        data = pd.read_csv(FETCH_URLS['adult'], header=None, na_values=' ?')\
            .dropna()
        data.rename(columns={data.columns[-1]: 'target'}, inplace=True)
        categorical_features = [1, 3, 5, 6, 7, 8, 9, 13]
        return data, categorical_features

    def fetch_abalone(self):
        """Download and transform the Abalone Data Set.

        https://archive.ics.uci.edu/ml/datasets/Abalone
        """
        data = pd.read_csv(FETCH_URLS['abalone'], header=None)
        data.rename(columns={data.columns[-1]: 'target'}, inplace=True)
        categorical_features = [0]
        return data, categorical_features

    def fetch_acute(self):
        """Download and transform the Acute Inflammations Data Set.

        https://archive.ics.uci.edu/ml/datasets/Acute+Inflammations
        """
        data = pd.read_csv(
            FETCH_URLS['acute'],
            header=None,
            sep='\t',
            decimal=',',
            encoding='UTF-16'
        )
        data['target'] = data[6].str[0] + data[7].str[0]
        data.drop(columns=[6, 7], inplace=True)
        categorical_features = list(range(1, 6))
        return data, categorical_features

    def fetch_annealing(self):
        """Download and transform the Annealing Data Set.

        https://archive.ics.uci.edu/ml/datasets/Annealing
        """
        data = pd.read_csv(FETCH_URLS['annealing'], header=None, na_values='?')

        # some features are dropped; they have too many missing values
        missing_feats = (data.isnull().sum(0) / data.shape[0]) < .1
        data = data.iloc[:, missing_feats.values]
        data[2].fillna(data[2].mode().squeeze(), inplace=True)

        data = data.T.reset_index(drop=True).T
        data.rename(columns={data.columns[-1]: 'target'}, inplace=True)

        categorical_features = [0, 1, 5, 9]
        return data, categorical_features

    def fetch_census(self):
        """Download and transform the Census-Income (KDD) Data Set.

        https://archive.ics.uci.edu/ml/datasets/Census-Income+%28KDD%29
        """
        data = pd.read_csv(FETCH_URLS['census'], header=None)

        categorical_features = list(range(1, 5))+list(range(6, 16)) +\
            list(range(19, 29))+list(range(30, 38))+[39]

        # some features are dropped; they have too many missing values
        cols_ids = [1, 6, 9, 13, 14, 20, 21, 29, 31, 37]
        categorical_features = np.argwhere(np.delete(
            data.rename(columns={k: f'nom_{k}' for k in categorical_features})
                .columns.astype('str').str.startswith('nom_'),
            cols_ids
        )).squeeze()
        data = data.drop(columns=cols_ids).T.reset_index(drop=True).T
        # some rows are dropped; they have rare missing values
        data = data.iloc[
            data.applymap(lambda x: x != ' Not in universe').all(1).values, :
        ]

        data.rename(columns={data.columns[-1]: 'target'}, inplace=True)
        return data, categorical_features

    def fetch_contraceptive(self):
        """Download and transform the Contraceptive Method Choice Data Set.

        https://archive.ics.uci.edu/ml/datasets/Contraceptive+Method+Choice
        """
        data = pd.read_csv(FETCH_URLS['contraceptive'], header=None)
        data.rename(columns={data.columns[-1]: 'target'}, inplace=True)
        categorical_features = [4, 5, 6, 8]
        return data, categorical_features

    def fetch_covertype(self):
        """Download and transform the Covertype Data Set.

        https://archive.ics.uci.edu/ml/datasets/Covertype
        """
        data = pd.read_csv(FETCH_URLS['covertype'], header=None)
        data.rename(columns={data.columns[-1]: 'target'}, inplace=True)
        wilderness_area = pd.Series(
            np.argmax(data.iloc[:, 10:14].values, axis=1),
            name=10
        )
        soil_type = pd.Series(
            np.argmax(data.iloc[:, 14:54].values, axis=1),
            name=11
        )
        data = data\
            .drop(columns=list(range(10, 54)))\
            .join(wilderness_area)\
            .join(soil_type)[list(range(0, 12))+['target']]
        categorical_features = [10, 11]
        return data, categorical_features

    def fetch_credit_approval(self):
        """Download and transform the Credit Approval Data Set.

        https://archive.ics.uci.edu/ml/datasets/Credit+Approval
        """
        data = pd.read_csv(
            FETCH_URLS['credit_approval'],
            header=None,
            na_values='?'
        ).dropna()
        data.rename(columns={data.columns[-1]: 'target'}, inplace=True)
        categorical_features = [0, 3, 4, 5, 6, 8, 9, 11, 12]
        return data, categorical_features

    def fetch_dermatology(self):
        """Download and transform the Dermatology Data Set.

        https://archive.ics.uci.edu/ml/datasets/Dermatology
        """
        data = pd.read_csv(
            FETCH_URLS['dermatology'],
            header=None,
            na_values='?'
        ).dropna()
        data.rename(columns={data.columns[-1]: 'target'}, inplace=True)
        categorical_features = list(range(data.shape[1]-1))
        categorical_features.remove(33)
        return data, categorical_features

    def fetch_echocardiogram(self):
        """Download and transform the Echocardiogram Data Set.

        https://archive.ics.uci.edu/ml/datasets/Echocardiogram
        """
        data = pd.read_csv(
            FETCH_URLS['echocardiogram'],
            header=None,
            error_bad_lines=False,
            warn_bad_lines=False,
            na_values='?'
        )
        data.drop(columns=[10, 11], inplace=True)
        data.dropna(inplace=True)
        data.rename(columns={data.columns[-1]: 'target'}, inplace=True)
        categorical_features = [1, 3]
        return data, categorical_features

    def fetch_flags(self):
        """Download and transform the Flags Data Set.

        https://archive.ics.uci.edu/ml/datasets/Flags
        """
        data = pd.read_csv(FETCH_URLS['flags'], header=None)
        target = data[6].rename('target')
        data = data.drop(columns=[0, 6])\
            .T.reset_index(drop=True).T.join(target)
        categorical_features = [
            0, 1, 4, 8, 9, 10, 11, 12, 13, 14, 15, 21, 22, 23, 24, 25, 26, 27
        ]
        return data, categorical_features

    def fetch_heart_disease(self):
        """Download and transform the Heart Disease Data Set.

        https://archive.ics.uci.edu/ml/datasets/Heart+Disease
        """
        data = pd.concat(
            [
                pd.read_csv(url, header=None, na_values='?')
                for url in FETCH_URLS['heart_disease']
            ],
            ignore_index=True
        ).drop(columns=[10, 11, 12]).dropna()
        data.rename(columns={data.columns[-1]: 'target'}, inplace=True)
        categorical_features = [1, 2, 5, 6, 8]
        return data, categorical_features

    def fetch_hepatitis(self):
        """Download and transform the Hepatitis Data Set.

        https://archive.ics.uci.edu/ml/datasets/Hepatitis
        """
        data = pd.read_csv(
            FETCH_URLS['hepatitis'],
            header=None,
            na_values='?'
        ).drop(columns=[15, 18]).dropna()
        target = data[0].rename('target')
        data = data.drop(columns=[0]).T.reset_index(drop=True).T.join(target)
        categorical_features = list(range(1, 13))+[16]
        return data, categorical_features

    def fetch_german_credit(self):
        """Download and transform the German Credit Data Set.

        https://archive.ics.uci.edu/ml/datasets/Statlog+%28German+Credit+Data%29
        """
        data = pd.read_csv(FETCH_URLS['german_credit'], header=None, sep=' ')
        data.rename(columns={data.columns[-1]: 'target'}, inplace=True)
        categorical_features = np.argwhere(
            data.iloc[0, :-1].apply(lambda x: str(x)[0] == 'A').values
        ).squeeze().tolist()
        return data, categorical_features

    def fetch_heart(self):
        """Download and transform the Heart Data Set.

        http://archive.ics.uci.edu/ml/datasets/statlog+(heart)
        """
        data = pd.read_csv(
            FETCH_URLS['heart'],
            header=None,
            delim_whitespace=True
        )
        data.rename(columns={data.columns[-1]: 'target'}, inplace=True)
        categorical_features = [1, 2, 5, 6, 8, 10, 12]
        return data, categorical_features

    def fetch_thyroid(self):
        """Download and transform the Thyroid Disease Data Set.
        Label 0 corresponds to no disease found.
        Label 1 corresponds to one or multiple diseases found.

        https://archive.ics.uci.edu/ml/datasets/Thyroid+Disease
        """
        data = pd.read_csv(FETCH_URLS['thyroid'], header=None, na_values='?')\
            .drop(columns=27).dropna().T.reset_index(drop=True).T
        data.rename(columns={data.columns[-1]: 'target'}, inplace=True)
        data['target'] = (
            data['target'].apply(lambda x: x.split('[')[0]) != '-'
        ).astype(int)
        categorical_features = [
            1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 18, 20,
            22, 24, 26, 27
        ]
        return data, categorical_features


class MulticlassDatasets(Datasets):
    """Class to download, transform and save multiclass datasets."""

    def fetch_first_order_theorem(self):
        """Download and transform the First Order Theorem Data Set.

        https://www.openml.org/d/1475
        """
        data = pd.read_csv(FETCH_URLS['first_order_theorem'])
        data.rename(columns={'Class': 'target'}, inplace=True)
        return data

    def fetch_gas_drift(self):
        """Download and transform the Gas Drift Data Set.

        https://www.openml.org/d/1476
        """
        data = pd.read_csv(FETCH_URLS['gas_drift'])
        data.rename(columns={'Class': 'target'}, inplace=True)
        return data

    def fetch_autouniv_au7(self):
        """Download and transform the AutoUniv au7 Data Set

        https://www.openml.org/d/1552
        """
        data = pd.read_csv(FETCH_URLS['autouniv_au7'])
        data.rename(columns={'Class': 'target'}, inplace=True)
        data.target = data.target\
            .apply(lambda x: x.replace('class', ''))\
            .astype(int)

        mask = (data.iloc[:, :-1].nunique() > 10).tolist()
        mask.append(True)
        data = data.loc[:, mask].copy()
        return data

    def fetch_autouniv_au4(self):
        """Download and transform the AutoUniv au4 Data Set

        https://www.openml.org/d/1548
        """
        data = pd.read_csv(FETCH_URLS['autouniv_au4'])
        data.rename(columns={'Class': 'target'}, inplace=True)
        data.target = data.target\
            .apply(lambda x: x.replace('class', ''))\
            .astype(int)

        mask = (data.iloc[:, :-1].nunique() > 10).tolist()
        mask.append(True)
        data = data.loc[:, mask].copy()
        return data

    def fetch_mice_protein(self):
        """Download and transform the Mice Protein Data Set

        https://www.openml.org/d/40966
        """
        data = pd.read_csv(FETCH_URLS['mice_protein'])
        data.rename(columns={'class': 'target'}, inplace=True)
        data.drop(columns=['MouseID'], inplace=True)
        data.replace('?', np.nan, inplace=True)

        mask = (data.iloc[:, :-1].nunique() > 10).tolist()
        mask.append(True)
        mask2 = data.isna().sum() < 10
        data = data.loc[:, mask & mask2].dropna().copy()

        data.iloc[:, :-1] = data.iloc[:, :-1].astype(float)

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)
        return data

    def fetch_steel_plates(self):
        """Download and transform the Steel Plates Fault Data Set.

        https://www.openml.org/d/40982
        """
        data = pd.read_csv(FETCH_URLS['steel_plates'])

        mask = (data.iloc[:, :-1].nunique() > 10).tolist()
        mask.append(True)
        data = data.loc[:, mask].copy()

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)
        return data

    def fetch_cardiotocography(self):
        """Download and transform the Cardiotocography Data Set.

        https://www.openml.org/d/1560
        """
        data = pd.read_csv(FETCH_URLS['cardiotocography'])
        data.rename(columns={'Class': 'target'}, inplace=True)

        mask = (data.iloc[:, :-1].nunique() > 10).tolist()
        mask.append(True)
        data = data.loc[:, mask].copy()

        return data

    def fetch_waveform(self):
        """Download and transform the Waveform Database Generator (version 2) Data Set.

        https://www.openml.org/d/60
        """
        data = pd.read_csv(FETCH_URLS['waveform'])
        data.rename(columns={'class': 'target'}, inplace=True)
        return data

    def fetch_volkert(self):
        """Download and transform the Volkert Data Set.

        https://www.openml.org/d/41166
        """
        data = pd.read_csv(FETCH_URLS['volkert'])
        data.rename(columns={'class': 'target'}, inplace=True)

        mask = (data.iloc[:, 1:].nunique() > 100).tolist()
        mask.insert(0, True)
        data = data.loc[:, mask].copy()
        return data

    def fetch_vehicle(self):
        """Download and transform the Vehicle Silhouettes Data Set.

        https://archive.ics.uci.edu/ml/datasets/Statlog+(Vehicle+Silhouettes)
        """
        data = pd.DataFrame()
        for letter in ascii_lowercase[0:9]:
            partial_data = pd.read_csv(
                urljoin(FETCH_URLS['vehicle'], 'xa%s.dat' % letter),
                header=None, delim_whitespace=True
            )
            partial_data = partial_data.rename(columns={18: 'target'})
            data = data.append(partial_data)

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)

        return data

    def fetch_asp_potassco(self):
        """Download and transform the ASP-POTASSCO Data Set.

        https://www.openml.org/d/41705
        """
        data = pd.read_csv(FETCH_URLS['asp_potassco'], na_values='?')
        data.dropna(inplace=True)
        data['target'] = data['algorithm']
        data.drop(columns=['instance_id', 'algorithm'], inplace=True)

        mask = (data.iloc[:, :-1].nunique() > 100).tolist()
        mask.append(True)
        data = data.loc[:, mask].copy()

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)

        return data

    def fetch_wine_quality(self):
        """Download and transform the Wine Quality Data Set.

        https://www.openml.org/d/40691
        """
        data = pd.read_csv(FETCH_URLS['wine_quality'])
        data.rename(columns={'class': 'target'}, inplace=True)
        return data

    def fetch_mfeat_zernike(self):
        """Download and transform the Multiple Features Dataset: Zernike Data Set.

        https://www.openml.org/d/22
        """
        data = pd.read_csv(FETCH_URLS['mfeat_zernike'])
        data.drop_duplicates(inplace=True)
        data.rename(columns={'class': 'target'}, inplace=True)
        return data

    def fetch_gesture_segmentation(self):
        """Download and transform the Gesture Phase Segmentation Data Set.

        https://www.openml.org/d/4538
        """
        data = pd.read_csv(FETCH_URLS['gesture_segmentation'])
        data.rename(columns={'Phase': 'target'}, inplace=True)

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)
        return data

    def fetch_texture(self):
        """Download and transform the Texture Data Set.

        https://www.openml.org/d/40499
        """
        data = pd.read_csv(FETCH_URLS['texture'])
        data.drop_duplicates(inplace=True)
        data.rename(columns={'Class': 'target'}, inplace=True)
        return data

    def fetch_usps(self):
        """Download and transform the USPS Data Set.

        https://www.openml.org/data/get_csv/19329737/usps.arff
        """
        data = pd.read_csv(FETCH_URLS['usps'])
        data.rename(columns={'int0': 'target'}, inplace=True)
        return data

    def fetch_japanese_vowels(self):
        """Download and transform the Japanese Vowels Data Set.

        https://www.openml.org/d/375
        """
        data = pd.read_csv(FETCH_URLS['japanese_vowels'])
        data.rename(columns={'speaker': 'target'}, inplace=True)
        data.drop(columns=['utterance', 'frame'], inplace=True)
        return data

    def fetch_pendigits(self):
        """Download and transform the Pen-Based Recognition of Handwritten
        Digits Data Set.

        https://www.openml.org/d/32
        """
        data = pd.read_csv(FETCH_URLS['pendigits'])
        data.rename(columns={'class': 'target'}, inplace=True)
        return data

    def fetch_image_segmentation(self):
        """Download and transform the Image Segmentation Data Set.

        https://www.openml.org/d/40984
        """
        data = pd.read_csv(FETCH_URLS['image_segmentation'])
        data.drop(columns=data.columns[:5], inplace=True)
        data.rename(columns={'class': 'target'}, inplace=True)

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)
        return data

    def fetch_baseball(self):
        """Download and transform the Baseball Hall of Fame Data Set.

        https://www.openml.org/d/185
        """
        data = pd.read_csv(FETCH_URLS['baseball'], na_values='?')
        data.drop(columns=['Player', 'Position'], inplace=True)
        data.rename(columns={'Hall_of_Fame': 'target'}, inplace=True)
        data.dropna(inplace=True)
        return data


class RemoteSensingDatasets(Datasets):
    """Class to download, transform and save remote sensing datasets."""

    def __init__(self, names='all', return_coords=False):
        self.names = names
        self.return_coords = return_coords

    def _load_gic_dataset(self, dataset_name):
        for url in FETCH_URLS[dataset_name]:
            r = requests.get(url, stream=True)
            content = loadmat(io.BytesIO(r.content))
            arr = np.array(list(content.values())[-1])
            arr = np.expand_dims(arr, -1) if arr.ndim == 2 else arr
            if self.return_coords and arr.shape[-1] != 1:
                indices = np.moveaxis(np.indices(arr.shape[:-1]), 0, -1)
                arr = np.insert(arr, [0, 0], indices, -1)
            yield arr

    def fetch_indian_pines(self):
        """Download and transform the Indian Pines Data Set. Label "0" means
        the pixel is not labelled. It is therefore dropped.

        http://www.ehu.eus/ccwintco/index.php/Hyperspectral_Remote_Sensing_Scenes#Indian_Pines
        """
        df = img_array_to_pandas(*self._load_gic_dataset('indian_pines'))
        return df[df.target != 0]

    def fetch_salinas(self):
        """Download and transform the Salinas Data Set. Label "0" means the pixel is
        not labelled. It is therefore dropped.

        http://www.ehu.eus/ccwintco/index.php/Hyperspectral_Remote_Sensing_Scenes#Salinas_scene
        """
        df = img_array_to_pandas(*self._load_gic_dataset('salinas'))
        return df[df.target != 0]

    def fetch_salinas_a(self):
        """Download and transform the Salinas-A Data Set. Label "0" means the pixel is
        not labelled. It is therefore dropped.

        http://www.ehu.eus/ccwintco/index.php/Hyperspectral_Remote_Sensing_Scenes#Salinas-A_scene
        """
        df = img_array_to_pandas(*self._load_gic_dataset('salinas_a'))
        return df[df.target != 0]

    def fetch_pavia_centre(self):
        """Download and transform the Pavia Centre Data Set. Label "0" means the pixel
        is not labelled. It is therefore dropped.

        http://www.ehu.eus/ccwintco/index.php/Hyperspectral_Remote_Sensing_Scenes#Pavia_Centre_scene
        """
        df = img_array_to_pandas(*self._load_gic_dataset('pavia_centre'))
        return df[df.target != 0]

    def fetch_pavia_university(self):
        """Download and transform the Pavia University Data Set. Label "0"
        means the pixel is not labelled. It is therefore dropped.

        http://www.ehu.eus/ccwintco/index.php/Hyperspectral_Remote_Sensing_Scenes#Pavia_University_scene
        """
        df = img_array_to_pandas(*self._load_gic_dataset('pavia_university'))
        return df[df.target != 0]

    def fetch_kennedy_space_center(self):
        """Download and transform the Kennedy Space Center Data Set. Label "0"
        means the pixel is not labelled. It is therefore dropped.

        http://www.ehu.eus/ccwintco/index.php/Hyperspectral_Remote_Sensing_Scenes#Kennedy_Space_Center_.28KSC.29
        """
        df = img_array_to_pandas(
            *self._load_gic_dataset('kennedy_space_center')
        )
        return df[df.target != 0]

    def fetch_botswana(self):
        """Download and transform the Botswana Data Set. Label "0" means the pixel is
        not labelled. It is therefore dropped.

        http://www.ehu.eus/ccwintco/index.php/Hyperspectral_Remote_Sensing_Scenes#Botswana
        """
        df = img_array_to_pandas(*self._load_gic_dataset('botswana'))
        return df[df.target != 0]



"""
Functions for visualization formatting or producing pre-formatted
visualizations.
"""

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.colors import rgb2hex, Normalize
from matplotlib.cm import ScalarMappable


def load_plt_sns_configs(font_size=8):
    """
    Load LaTeX style configurations for Matplotlib/Seaborn
    Visualizations.
    """
    sns.set_style('whitegrid')
    tex_fonts = {
        # Use LaTeX to write all text
        "text.usetex": True,
        "font.family": "serif",
        # Use 10pt font in plots, to match 10pt font in document
        "axes.labelsize": (10/8)*font_size,
        "font.size": (10/8)*font_size,
        # Make the legend/label fonts a little smaller
        "legend.fontsize": font_size,
        "xtick.labelsize": font_size,
        "ytick.labelsize": font_size,
        # Subplots size/shape
        "figure.subplot.left": .098,
        "figure.subplot.right": .938,
        "figure.subplot.bottom": .12,
        "figure.subplot.top": .944,
        "figure.subplot.wspace": .071,
        "figure.subplot.hspace": .2
    }
    plt.rcParams.update(tex_fonts)


def val_to_color(col, cmap='RdYlBu_r'):
    """
    Converts a column of values to hex-type colors.

    Parameters
    ----------
    col : array-like of shape (n_samples,)
        Values to convert to hex-type color code

    cmap : str or `~matplotlib.colors.Colormap`
        The colormap used to map normalized data values to RGBA colors

    Returns
    -------
    colors : array-like of shape (n_samples,)
        Array with hex values as string type.
    """
    norm = Normalize(vmin=col.min(), vmax=col.max(), clip=True)
    mapper = ScalarMappable(norm=norm, cmap=cmap)
    rgba = mapper.to_rgba(col)

    return np.apply_along_axis(rgb2hex, 1, rgba)



"""
This submodule contains various performance metrics/scorers that are not
included in scikit-learn's scorers' dictionary. Additionally, an expanded
dictionary of scorers (as compared with scikit-learn's) is also provided.

Parts of this code was taken from the `utils.py` script from AlgoWit's
publications repo, to which I have also contributed.

Link to related repo: https://github.com/AlgoWit/publications
"""
from ._metrics import (
    SCORERS,
    geometric_mean_score_macro,
    ALScorer,
    area_under_learning_curve,
    data_utilization_rate
)

__all__ = [
    'SCORERS',
    'geometric_mean_score_macro',
    'area_under_learning_curve',
    'data_utilization_rate',
    'ALScorer'
]



import os
import codecs
from setuptools import find_packages, setup

ver_file = os.path.join('research', '_version.py')
with open(ver_file) as f:
    exec(f.read())

with open("requirements.txt") as reqs:
    REQUIREMENTS = [reqs.readlines()]

with open("requirements.dev.txt") as dev_reqs:
    REQUIREMENTS_DEV = [dev_reqs.readlines()]

MAINTAINER = 'J. Fonseca'
MAINTAINER_EMAIL = 'jpfonseca@novaims.unl.pt'
URL = 'https://github.com/joaopfonseca/ml-research'
VERSION = __version__
SHORT_DESCRIPTION = 'Implementation of Machine Learning algorithms, experiments and utilities.'
with codecs.open('README.rst', encoding='utf-8-sig') as f:
    LONG_DESCRIPTION = f.read()
LICENSE = 'MIT'
CLASSIFIERS = ['Intended Audience :: Science/Research',
               'Intended Audience :: Developers',
               'License :: OSI Approved',
               'Programming Language :: Python',
               'Topic :: Software Development',
               'Topic :: Scientific/Engineering',
               'Operating System :: Microsoft :: Windows',
               'Operating System :: POSIX',
               'Operating System :: Unix',
               'Operating System :: MacOS',
               'Programming Language :: Python :: 3.7',
               'Programming Language :: Python :: 3.8',
               'Programming Language :: Python :: 3.9']

setup(
    name='ml-research',
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    url=URL,
    download_url=URL,
    version=VERSION,
    description=SHORT_DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    license=LICENSE,
    classifiers=CLASSIFIERS,
    packages=find_packages(),
    install_requires=REQUIREMENTS,
    extras_require={'dev': REQUIREMENTS_DEV}
)
